package JavaCode;

import java.io.*;
import java.util.Collections;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.StringTokenizer;
import java.util.Vector;

/*
SLIDER: Local Short Reads DNA Sequence Alignment
Beta Release; Ver 0.1, 22th Mar. 2008
By: Nawar Malhis
Canada's Michael Smith Genome Sciences Centre
*/

public class OligoFile {
	
	static private final int buffSize = 256000;
	
	static public final int iDfns = 0;
	static public final int iDfn  = 0;
	
	static public final int F = 0;
	static public final int S = 1;
	static public final int W = 0;
	static public final int R = 1;
	static public boolean display   = false;

	public String  vName;           // vertual file name
	public String  rName;           // reference name, used in ACE
	public boolean read;            // open for read
	public int     dfn;             // data file number
	public int     dfns;            // data file number
	public long    oPointer;        // oligo Pointer in total
	public Vector<Oligo> bunkPet[][]; // [side][rc]
	// public Vector<PET> pets;
	public int     nopet;
	
	// meta data parameters
	private boolean ow;
	public int    numberOfBACs;
	public String[] BACs;
	public long   minID;
	public long   numberOfOligos;
	public int    numberOfFiles;
	public long   numberOfUQ;
	public long   numberOfUD;
	public double weightOfAT;
	public double weightOfCG;
	
	public int sortCounter;
	int[] l;
	long idx;
	
	Vector<Oligo> bunk;
	PriorityQueue<OligoFileTop> pQueue;
	OligoFile trg;
	
	FileInputStream fi; 
	BufferedInputStream bi;
	DataInputStream strIn;
	
	FileOutputStream fw;
	BufferedOutputStream bw;
	DataOutputStream strOut;
	
	FileInputStream[]     sfi; 
	BufferedInputStream[] sbi;
	DataInputStream[]  sstrIn;

	public class OligoFileTop implements Comparable<OligoFileTop>{
		// public boolean ok;
		public Oligo ol;
		public int   id;
		
		public OligoFileTop() throws Exception {
			ol = new Oligo();
			id = -1;
			// ok = true;
		}
		
		public int compareTo(OligoFileTop oft) {
			return ol.compareTo(oft.ol);
		}
		
		public void loadOligo(Oligo o) {
			ol.d[0]   = o.d[0];
			ol.d[1]   = o.d[1];
			ol.ch     = o.ch;
			ol.start  = o.start;
			ol.id1    = o.id1;
			ol.size   = o.size;
			ol.flg    = o.flg;
			ol.weight = o.weight;
	    	ol.u0     = o.u0;	    	
		}
	}

    static void sort_file(String fin, int st) throws Exception { 
    	int oldSt = Oligo.sortIndex;
    	Oligo.sortIndex = st;
    	String fout = fin + ".s";
    	if (Oligo.sortIndex == Oligo.SEQ) {
    		fout = fout + ".sq";
    	} else if (Oligo.sortIndex == Oligo.ID) {
    		fout = fout + ".id";
    	} else if (Oligo.sortIndex == Oligo.WEIGHT) {
    		fout = fout + ".w";
    	} else if (Oligo.sortIndex == Oligo.LOCATION) {
    		fout = fout + ".l";
    	}
		OligoFile ofile = new OligoFile(fin, "", OligoFile.R);
		if (ofile.numberOfOligos > 0) {
			ofile.sort(fout);
		}
		ofile.close();
		ofile = null;
    	Oligo.sortIndex = oldSt;
    }

	public OligoFile() {
		System.err.println("Error: must use a String constructore to create directory\n object not valid");
	}	
	public OligoFile(String fName, String refName, int fType) throws Exception {
		vName         = fName;
		rName         = refName;
		dfn           = iDfn;
		dfns          = iDfns;
		oPointer      = 0;
		numberOfFiles = 1;
		
		sortCounter   = 0;
		l = new int[2];
		nopet = 0;
    	bunk = new Vector<Oligo>();
    	// pets = new Vector<PET>();
		if (fType == R) {	
			openRead();
		} else {
			openWrite();
			if(!(new File(vName)).mkdir()) {
				System.err.println("Error creating the directory " + vName);
			}
		}
	}
	//*
	public OligoFile(String fullName) throws Exception {
		int f;
		int k;
		int ID;
		int next;
		boolean EOF;
		String outOl  = Parameters.pathData + "s.ol0/";
		String outPrb = Parameters.pathData + "/s.prb";
		FileOutputStream osPrb    = new FileOutputStream(outPrb);
		DataOutputStream dataout2 = new DataOutputStream(osPrb);
		DataInputStream strInPrb;
		OligoFile[] inOF;
		int numberOfDirs = 0;
		String[] inFiles;
		File fl = new File(Parameters.pathParallel);
		File[] dirs = fl.listFiles();

		for (int dd = 0; dd < dirs.length; dd++) {
			if (dirs[dd].isDirectory()) {
				numberOfDirs++;
			}
		}
		inFiles = new String[numberOfDirs];
		f = 0;
		for (int dd = 0; dd < dirs.length; dd++) {
			if (dirs[dd].isDirectory()) {
				inFiles[f] = new String(dirs[dd]+ "");
				f++;
			}
		}
		inOF  = new OligoFile[numberOfDirs];
		int[][] shift_array = new int[numberOfDirs + 1][3];
		
		for (int i = 0; i < inFiles.length; i++) {
			System.err.println(i + "\t" + inFiles[i] + "/s.prb");
			strInPrb = new DataInputStream(new BufferedInputStream(new FileInputStream(inFiles[i] + "/s.prb"))) ;
			k = 0;
			ID = 0;
			EOF = false;
			while (!EOF) {
				try {
					ID = strInPrb.readInt();
					if (k == 0) {
						if (i == 0) {
							shift_array[i][0] = ID;
							shift_array[i][2] = 0;
						} else {
							shift_array[i][0] = shift_array[i - 1][1] - 1;
							shift_array[i][2] = ID - shift_array[i - 1][1] + 1;
						}
						dataout2.writeInt(shift_array[i][0]);
					} else {
						if (i == 0) {
							dataout2.writeInt(ID);
						} else {
							dataout2.writeInt(ID - shift_array[i][2]);
						}
					}

					Byte B1 = strInPrb.readByte();
					dataout2.writeByte(B1);
					for (int ii = 0; ii < Parameters.oMxSize; ii++) {
						for (int jj = 0; jj < 4; jj++) {
							Byte B2 = strInPrb.readByte();
							dataout2.writeByte(B2);
						}
					}
					k++;
				} catch (EOFException e) {
					EOF = true;
				}
			}
			if (i == 0) {
				shift_array[i][1] = ID;
			} else {
				shift_array[i][1] = ID - shift_array[i][2];
			}
			strInPrb.close();
		} // ===============================================
		dataout2.close();

		// [numberOfDirs + 1][3]
		for (int i = 0; i < numberOfDirs; i++) {
			for (int j = 0; j < 3; j++) {
				System.err.print(shift_array[i][j] + "\t");
			}
			System.err.println();
		}
		//*
		OligoFile CompleteOligoFile = new OligoFile(outOl, "", OligoFile.W);

		PriorityQueue<OligoFileTop> pQueue = new PriorityQueue<OligoFileTop>();
		OligoFileTop oft;

		for (int i = 0 ; i < inOF.length; i++)
		{
			inOF[i] = new OligoFile(inFiles[i], "", OligoFile.R);
			oft = new OligoFileTop();
			Oligo ol = new Oligo();
			inOF[i].readOligo(ol);
			oft.ol = ol;
			oft.id = i;
			pQueue.add(oft);
		}
		
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			if (oft.id != 0) {
				oft.ol.id1 = oft.ol.id1 - shift_array[oft.id][2];
			}
			CompleteOligoFile.writeOligo(oft.ol);
			Oligo ol = new Oligo();
			if (inOF[next].readOligo(ol)) {	
				oft.ol = ol;
				oft.id = next;
				pQueue.add(oft);
			}
		}

		CompleteOligoFile.close();
		CompleteOligoFile.writeMeta();
		//*/
	}		

	
    public void finalize() throws Exception {
	}
	public void close() throws Exception {
    	if (read) {
    		if (idx > 0)
      			closeFileRead();
    	} else {
    		if(ow)
    			closeFileWrite();
    		writeMeta();
    	}
    	bunk.clear();
    	bunk.trimToSize();
    	System.gc();
	}
    
	public void openRead() throws Exception {
		read = true;
		idx  = 0;
		readMeta();
	}
	public void openWrite() throws Exception {
		read = false;
		ow = false;
		clearMeta();
	}

    public void readMeta() throws Exception {
    	// System.err.println("readMeta\t" + vName);
		StringTokenizer tokens;
		String line;
		BufferedReader reader = new BufferedReader(new FileReader(vName + "/meta.txt"));
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);	
		if (tokens.countTokens() >= 2) {
			tokens.nextToken();
			rName      = tokens.nextToken();
		}
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		minID          = new Long(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfOligos = new Long(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfUQ     = new Long(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfUD     = new Long(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		weightOfAT     = new Double(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		weightOfCG     = new Double(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfFiles  = new Integer(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);	
		tokens.nextToken();
		numberOfBACs   = new Integer(tokens.nextToken());

		if (numberOfBACs > 0) {
			BACs = new String[numberOfBACs];
			for (int i = 0; i < numberOfBACs; i++) {
				BACs[i] = reader.readLine();
			}
		}
		reader.close();
    }  
    public void writeMeta() throws Exception {
    	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/meta.txt"));
    	// System.err.println(vName + "/meta.txt,\tminID\t" + (minID-1));
    	writer.write("Ref:\t" + rName); // numberOfIDs
    	writer.write("\nminID:\t"      + minID);
    	writer.write("\nnumberOfOligos:\t" + numberOfOligos);
    	writer.write("\nnumberOfUQ:\t" + numberOfUQ);
    	writer.write("\nnumberOfUD:\t" + numberOfUD);
    	writer.write("\nweightOfAT:\t" + weightOfAT);
    	writer.write("\nweightOfCG:\t" + weightOfCG);
    	writer.write("\ndfn:\t" + dfn);
    	writer.write("\nnumberOfBACs:\t" + numberOfBACs + "\n");
		for (int i = 0; i < numberOfBACs; i++) {
			writer.write(BACs[i] + "\n");
		}
    	writer.close();
    }
    public void clearMeta() {
    	minID          = Parameters.maxId;
    	numberOfOligos = 0;
    	weightOfAT     = 0;
    	weightOfCG     = 0;
    	numberOfUQ     = 0;
    	numberOfUD     = 0;
    	numberOfFiles  = 0;
    }
    public void displayMeta() {
		System.out.println( "\n\n" + 
				"Meta data for\t" + vName + "\n" +
				"Number of files .................:\t" + numberOfFiles + "\n" +
				"Number of Oligos ................:\t" + numberOfOligos + "\n" +
				"Minimum IDs .....................:\t" + minID + "\n" + 
				"Weight of Oligos ................:\t" + (long) (weightOfAT  + weightOfCG) + "\n" + 
				"Average Oligo Weight ............:\t" + Parameters.trim((weightOfAT  + weightOfCG) / (Parameters.maxId - minID)) + "\n" + 
				"Number of UQ ....................:\t" + numberOfUQ + "\n" + 
				"Number of UD ....................:\t" + numberOfUD + "\n" + 
				"Multiplication ratio of UD Oligos:\t" + ((float) numberOfUD / ((float) ((Parameters.maxId - minID) - numberOfUQ))) + "\n" +
				"Weight of AT ....................:\t" + (long) weightOfAT + "\n" + 
				"Weight of CG ....................:\t" + (long) weightOfCG + "\n" +
				"Ratio AT/CG .....................:\t" + Parameters.trim(((float) weightOfAT / (float) weightOfCG)) +
				"\n");
    }
    
    public void info(int fStart) throws Exception {
    	int nid = (int) (Parameters.maxId + 1 - minID);
    	long[] histo = new long[Parameters.dbOligoSize + 1];
    	int hi;
    	byte[][] cnt;
    	byte[][] cntLQ;
    	byte[][] cntCrsp;
    	byte[][] cntUM;
    	byte[][] cntUMLQ;
    	byte[][] cntUMCrsp;
    	long countMP    = 0;
    	long countUM    = 0;
    	long countUQ    = 0;
    	long countRC    = 0;
    	long countLQ    = 0;
    	long countUMRC  = 0;
    	long countUMS0  = 0;
    	long countMPS0  = 0;
    	long countLQS0  = 0;
    	long countS0    = 0;
    	long countUQS0  = 0;
    	long countTotal = 0;
    	long countPET   = 0;
    	float weight  = 0;
    	int cntId;
    	
		cnt       = new byte[2][(int) nid];
		cntLQ     = new byte[2][(int) nid];
		cntCrsp   = new byte[2][(int) nid];
		cntUM     = new byte[2][(int) nid];
		cntUMLQ   = new byte[2][(int) nid];
		cntUMCrsp = new byte[2][(int) nid];
		for(int ii = 0; ii < nid; ii++) {
			cnt[0][ii]       = cnt[1][ii]       = 0;
			cntLQ[0][ii]     = cntLQ[1][ii]     = 0;
			cntCrsp[0][ii]   = cntCrsp[1][ii]   = 0;
			cntUM[0][ii]     = cntUM[1][ii]     = 0;
			cntUMLQ[0][ii]   = cntUMLQ[1][ii]   = 0;
			cntUMCrsp[0][ii] = cntUMCrsp[1][ii] = 0;
		}
		System.err.println("Stat\t" + vName);
    	if (read) {
   			openFileRead(vName + "/file" + fStart);
   			displayMeta();
    		dfn = fStart;
        	Oligo ol = new Oligo();
			idx = 0;
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				readOligo(ol, "/file", Parameters.fSize);
				if(ol.isPET())
					countPET++;
	   			countTotal++;
				cntId = Parameters.maxId - ol.id1;
				if (ol.u0 >= 0) {
					if (ol.u0 < Parameters.dbOligoSize)
						hi = ol.u0;
					else 
						hi = Parameters.dbOligoSize;
					histo[hi]++;
				} else {
					/*
					countNegativ++;
					if ((countNegativ % 1000) == 0)
						System.err.println(countNegativ + "\t" + ol);
					Thread.sleep(1000);
					//*/
					// System.err.println("countNegativ\t" + ol);
				}
				if (cntId < nid) {
					cnt[ol.getSide()][cntId] = 1;
					if (ol.isLQ()) {
						cntLQ[ol.getSide()][cntId] = 1;
					}
					if (ol.isUQ()) {
						cntCrsp[ol.getSide()][cntId] = 1;
					}
					if (ol.isUM()) {
						cntUM[ol.getSide()][cntId] = 1;
						if (ol.isLQ()) {
							cntUMLQ[ol.getSide()][cntId] = 1;
						}
						if (ol.isUQ()) {
							cntUMCrsp[ol.getSide()][cntId] = 1;
						}
					}
				} else {
					System.err.println(" --------------- id out of range" + cntId);
				}

				if(ol.isMP()) {
					countMP++;
					if(ol.isSide0())
						countMPS0++;
				}
				if(ol.isLQ()) {
					countLQ++;
					if(ol.isSide0())
						countLQS0++;
				}
				if(ol.isRC()) {
					countRC++;
					if(ol.isUM())
						countUMRC++;
				}
				if(ol.isUM()) {
					countUM++;
					if(ol.isSide0())
						countUMS0++;
				}
				if(ol.isSide0())
					countS0++;
				if(ol.isUQ()) {
					countUQ++;
					if(ol.isSide0())
						countUQS0++;
				}
				weight += ol.weight;
				idx++;
			}		
    	} else {
			System.err.println(" File open for write");
		}
    	int cntRead  = 0;
    	int cntS0    = 0;
    	int cntS1    = 0;
    	int cntS0oS1 = 0;
    	int cntS0aS1 = 0;

    	int cntLQ1   = 0;
    	int cntLQ2   = 0;
    	int cntLQ1o2 = 0;
    	int cntLQ1a2 = 0;

    	int cntCrsp1   = 0;
    	int cntCrsp2   = 0;
    	int cntCrsp1o2 = 0;
    	int cntCrsp1a2 = 0;

    	int cntUM1   = 0;
    	int cntUM2   = 0;
    	int cntUM1o2 = 0;
    	int cntUM1a2 = 0;

    	int cntUMLQ1   = 0;
    	int cntUMLQ2   = 0;
    	int cntUMLQ1o2 = 0;
    	int cntUMLQ1a2 = 0;

    	int cntUMCrsp1   = 0;
    	int cntUMCrsp2   = 0;
    	int cntUMCrsp1o2 = 0;
    	int cntUMCrsp1a2 = 0;

    	System.err.println("\nInfo:");
     	cntId = 0;
    	for(int ii = 0; ii < nid; ii++) {
    		if(cnt[0][ii] > 0 || cnt[1][ii] > 0) {
    			cntS0oS1++;
    		}
    		if(cnt[0][ii] > 0) {
    			cntS0++;
    			cntRead++;
        		if(cnt[1][ii] > 0) {
        			cntS0aS1++;
        		}
    		}
    		if(cnt[1][ii] > 0) {
    			cntRead++;
    			cntS1++;
    		}
    		
    		if(cntCrsp[0][ii] > 0 || cntCrsp[1][ii] > 0) {
    			cntCrsp1o2++;
    		}
    		if(cntCrsp[0][ii] > 0) {
    			cntCrsp1++;
        		if(cntCrsp[1][ii] > 0) {
        			cntCrsp1a2++;
        		}
    		}
    		if(cntCrsp[1][ii] > 0) {
    			cntCrsp2++;
    		}

    		if(cntLQ[0][ii] > 0 || cntLQ[1][ii] > 0) {
    			cntLQ1o2++;
    		}
    		if(cntLQ[0][ii] > 0) {
    			cntLQ1++;
        		if(cntLQ[1][ii] > 0) {
        			cntLQ1a2++;
        		}
    		}
    		if(cntLQ[1][ii] > 0) {
    			cntLQ2++;
    		}

    		if(cntUM[0][ii] > 0 || cntUM[1][ii] > 0) {
    			cntUM1o2++;
    		}
    		if(cntUM[0][ii] > 0) {
    			cntUM1++;
        		if(cntUM[1][ii] > 0) {
        			cntUM1a2++;
        		}
    		}
    		if(cntUM[1][ii] > 0) {
    			cntUM2++;
    		}

    		if(cntUMCrsp[0][ii] > 0 || cntUMCrsp[1][ii] > 0) {
    			cntUMCrsp1o2++;
    		}
    		if(cntUMCrsp[0][ii] > 0) {
    			cntUMCrsp1++;
        		if(cntUMCrsp[1][ii] > 0) {
        			cntUMCrsp1a2++;
        		}
    		}
    		if(cntUMCrsp[1][ii] > 0) {
    			cntUMCrsp2++;
    		}

    		if(cntUMLQ[0][ii] > 0 || cntUMLQ[1][ii] > 0) {
    			cntUMLQ1o2++;
    		}
    		if(cntUMLQ[0][ii] > 0) {
    			cntUMLQ1++;
        		if(cntUMLQ[1][ii] > 0) {
        			cntUMLQ1a2++;
        		}
    		}
    		if(cntUMLQ[1][ii] > 0) {
    			cntUMLQ2++;
    		}
    	}
		// System.err.println("countIDs :\t" + cntId);
		System.err.println("S0 +  S1 ..:\t" + cntRead);
		System.err.println("S0 ........:\t" + cntS0 + "\t" + Parameters.trim(((float) cntS0 * 100 / cntRead)) + "% of reads");
		System.err.println("S1 ........:\t" + cntS1 + "\t" + Parameters.trim(((float) cntS1 * 100 / cntRead)) + "% of reads");
		System.err.println("S0 && S1 ..:\t" + cntS0aS1 + "\t" + Parameters.trim(((float) cntS0aS1 * 200 / cntRead)) + "% of reads");
		System.err.println("PETs ......:\t" + (countPET/2.0) + "\t" + Parameters.trim(((float) countPET * 100 / cntRead)) + "% of reads");
		System.err.println("S0 || S1 ..:\t" + cntS0oS1);
		
		System.err.println("\nCrisp0 + 1  :\t" + (cntCrsp1 + cntCrsp2)); 
		System.err.println("Crisp0 .......:\t" + cntCrsp1 + "\t" + Parameters.trim(((float) cntCrsp1 * 100 / (cntCrsp1 + cntCrsp2))) + "% of Crisps"); 
		System.err.println("Crisp1 .......:\t" + cntCrsp2 + "\t" + Parameters.trim(((float) cntCrsp2 * 100 / (cntCrsp1 + cntCrsp2))) + "% of Crisps");
		System.err.println("Crisp0 && 1 ..:\t" + cntCrsp1a2 + "\t" + Parameters.trim(((float) cntCrsp1a2 * 200 / (cntCrsp1 + cntCrsp2))) + "% of Crisps");
		System.err.println("Crisp0 || 1 ..:\t" + cntCrsp1o2);

		System.err.println("\nLQ0 + 1  :\t" + (cntLQ1 + cntLQ2)); 
		System.err.println("LQ0 .......:\t" + cntLQ1 + "\t" + Parameters.trim(((float) cntLQ1 * 100 / (cntLQ1 + cntLQ2))) + "% of LQs"); 
		System.err.println("LQ1 .......:\t" + cntLQ2 + "\t" + Parameters.trim(((float) cntLQ2 * 100 / (cntLQ1 + cntLQ2))) + "% of LQs");
		System.err.println("LQ0 && 1 ..:\t" + cntLQ1a2 + "\t" + Parameters.trim(((float) cntLQ1a2 * 200 / (cntLQ1 + cntLQ2))) + "% of LQs");
		System.err.println("LQ0 || 1 ..:\t" + cntLQ1o2);

		System.err.println("\nUM0 + UM1  :\t" + (cntUM1 + cntUM2)); 
		System.err.println("UM0 .......:\t" + cntUM1 + "\t" + Parameters.trim(((float) cntUM1 * 100 / (cntUM1 + cntUM2))) + "% of UMs"); 
		System.err.println("UM1 .......:\t" + cntUM2 + "\t" + Parameters.trim(((float) cntUM2 * 100 / (cntUM1 + cntUM2))) + "% of UMs");
		System.err.println("UM0 && 1 ..:\t" + cntUM1a2 + "\t" + Parameters.trim(((float) cntUM1a2 * 200 / (cntUM1 + cntUM2))) + "% of UMs");
		System.err.println("UM0 || UM1 :\t" + cntUM1o2);

		System.err.println("\nUMCrisp0 + 1  :\t" + (cntUMCrsp1 + cntUMCrsp2)); 
		System.err.println("UMCrisp0 .......:\t" + cntUMCrsp1 + "\t" + Parameters.trim(((float) cntUMCrsp1 * 100 / (cntUMCrsp1 + cntUMCrsp2))) + "% of UM Crisps"); 
		System.err.println("UMCrisp1 .......:\t" + cntUMCrsp2 + "\t" + Parameters.trim(((float) cntUMCrsp2 * 100 / (cntUMCrsp1 + cntUMCrsp2))) + "% of UM Crisps");
		System.err.println("UMCrisp0 && 1 ..:\t" + cntUMCrsp1a2 + "\t" + Parameters.trim(((float) cntUMCrsp1a2 * 200 / (cntUMCrsp1 + cntUMCrsp2))) + "% of UM Crisps");
		System.err.println("UMCrisp0 || 1 ..:\t" + cntUMCrsp1o2);

		System.err.println("\nUMLQ0 + 1  :\t" + (cntUMLQ1 + cntUMLQ2)); 
		System.err.println("UMLQ0 .......:\t" + cntUMLQ1 + "\t" + Parameters.trim(((float) cntUMLQ1 * 100 / (cntUMLQ1 + cntUMLQ2))) + "% of UM LQs"); 
		System.err.println("UMLQ1 .......:\t" + cntUMLQ2 + "\t" + Parameters.trim(((float) cntUMLQ2 * 100 / (cntUMLQ1 + cntUMLQ2))) + "% of UM LQs");
		System.err.println("UMLQ0 && 1 ..:\t" + cntUMLQ1a2 + "\t" + Parameters.trim(((float) cntUMLQ1a2 * 200 / (cntUMLQ1 + cntUMLQ2))) + "% of UM LQs");
		System.err.println("UMLQ0 || 1 ..:\t" + cntUMLQ1o2);


    	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/histogramU0.txt"));
    	for (int i = 0; i < histo.length; i++) {
    		writer.write(i + "\t" + histo[i] + "\n");
    	}
    	writer.close();
		System.err.println("countTotal  :\t" + countTotal);
		System.err.println("countMP ....:\t" + countMP + "\t" + Parameters.trim(((float) countMP * 100 / countTotal)) + "%");
		System.err.println("countMPS0 ..:\t" + countMPS0 + "\t" + Parameters.trim((float) countMPS0 * 100 / countTotal) + "%");
		System.err.println("countLQ ....:\t" + countLQ + "\t" + Parameters.trim(((float) countLQ * 100 / countTotal)) + "%");
		System.err.println("countLQS0 ..:\t" + countLQS0 + "\t" + Parameters.trim(((float) countLQS0 * 100 / countTotal)) + "%");
		System.err.println("countS0 ....:\t" + countS0 + "\t" + Parameters.trim((float) countS0 * 100 / countTotal) + "%");
		System.err.println("countRC ....:\t" + countRC + "\t" + Parameters.trim((float) countRC * 100 / countTotal) + "%");
		System.err.println("countUM ....:\t" + countUM + "\t" + Parameters.trim((float) countUM * 100 / countTotal) + "%");
		System.err.println("countUMRC ..:\t" + countUMRC + "\t" + Parameters.trim((float) countUMRC * 100 / countTotal) + "%");
		System.err.println("countUMS0 ..:\t" + countUMS0 + "\t" + Parameters.trim((float) countUMS0 * 100 / countTotal) + "%");
		System.err.println("countUMS1 ..:\t" + (countUM - countUMS0) + "\t" + Parameters.trim((float) (countUM - countUMS0) * 100  / countTotal) + "%");
		System.err.println("countCrisp  :\t" + countUQ + "\t" + Parameters.trim((float) countUQ * 100 / countTotal) + "%");
		System.err.println("countCrispU0:\t" + countUQS0 + "\t" + Parameters.trim((float) countUQS0 * 100 / countTotal) + "%");
		System.err.println("Weight .....:\t" + weight + "\t" + Parameters.trim((float) weight * 100 / countTotal));
    }
    /*
    public void splitPET(String of, int fStart) throws Exception {
    	int insertions = 0;
    	int totalCount = 0;
    	int idBak;
    	int mpCount = 0;
    	int rc;
    	PET pt = new PET();
		bunkPet       = new Vector[2][2]; // [side][rc]
		bunkPet[0][0] = new Vector<Oligo>(); 
		bunkPet[0][1] = new Vector<Oligo>(); 
		bunkPet[1][0] = new Vector<Oligo>(); 
		bunkPet[1][1] = new Vector<Oligo>(); 

    	OligoFile ofile = new OligoFile(of, "", OligoFile.W);
		System.err.println("splitPET\t" + vName );
    	if (read) {
   			openFileRead(vName + "/file" + fStart);
    		dfn = fStart;
        	Oligo ol = new Oligo();
        	readOligo(ol);
			idx = 1;
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				bunkPet[0][0].clear();
				bunkPet[0][1].clear();
				bunkPet[1][0].clear();
				bunkPet[1][1].clear();
				mpCount = 0;
				idBak = ol.id1;
				while (ol.id1 == idBak && idx < numberOfOligos) {
					idx++;
					if(ol.isMP())
						mpCount++;
		        	if (ol.u0 <= Parameters.dbOligoSize) { // [side][rc]
		        		if(ol.isRC()) {
		        			rc = 1;
						} else {
							rc = 0;
							
						}
		        		bunkPet[ol.getSide()][rc].add(new Oligo(ol));
		        	}
					readOligo(ol);
				}
				if (ol.id1 == idBak && idx == numberOfOligos) {
					idx++;
					if(ol.isMP())
						mpCount++;
		        	if (ol.u0 <= Parameters.dbOligoSize) {
		        		if(ol.isRC())
		        			rc = 1;
		        		else
		        			rc = 0;
		        		bunkPet[ol.getSide()][rc].add(new Oligo(ol));
		        	}
				} 
				findPets();
				Collections.sort(pets);
				Iterator<PET> it0 = pets.iterator();
				while (it0.hasNext()) {
					pt = new PET((PET) it0.next ());
					System.err.println(pt.score);
				}
				Thread.sleep(1000);
			}	
    	} else {
			System.err.println(" File open for write");
		}
    	System.err.println("TotalCount:\t" + totalCount);
    	System.err.println("Insertions:\t" + insertions);
		ofile.saveSBunk(Oligo.SEQ);
		ofile.dfn = ofile.dfns;
		ofile.close();
		ofile = new OligoFile(of, "", OligoFile.R);
		ofile.joinSort();
		ofile.close();
    	
		displayMeta();
    }
    
    public void findPets() throws Exception {
		int pSize;
    	int rc0;
    	int rc1;
    	int oldSI;
    	Oligo[] olOut = new Oligo[2];
    	for(int i = 0; i < 2; i++) {
    		olOut[i] = new Oligo();
    	}
    	pets.clear();
		for (int l = 0; l < 2; l++) {// ------------------------------------------------------------ finding best pet
			if (l == 0) {
				rc0 = 0;
				rc1 = 1; 
			} else {
				rc0 = 1;
				rc1 = 0; 					
			}
			if(!(bunkPet[0][rc0].isEmpty() || bunkPet[0][rc0].isEmpty())) {
				// System.err.println("totalPETs:\t" + totalPETs + "\tsd0:\t" + sd[0] + ",\tsd1:\t" + sd[1] + "\tmpCount:\t" + mpCount);
				oldSI = Oligo.sortIndex;
				Oligo.sortIndex = Oligo.LOCATION;
				Collections.sort(bunkPet[0][rc0]);
				Collections.sort(bunkPet[1][rc1]);
				Oligo.sortIndex = oldSI;
				Iterator<Oligo> it0 = bunkPet[0][rc0].iterator();
				while (it0.hasNext()) {
					Iterator<Oligo> it1 = bunkPet[1][rc1].iterator();
					olOut[0] = new Oligo((Oligo) it0.next ());
			        while (it1.hasNext()) {
						olOut[1] = new Oligo((Oligo) it1.next ());
			        	if(olOut[0].ch == olOut[1].ch) {
			        		pSize = Math.abs(olOut[1].start - olOut[0].start);
			        		if (pSize < Parameters.maxPetSpam) {
				        		// totalCount++;
				        		olOut[0].setPET(true);
				        		olOut[1].setPET(true);
				        		if(olOut[0].u0 < olOut[1].u0)
				        			olOut[1].u0 = olOut[0].u0;
				        		else
				        			olOut[0].u0 = olOut[1].u0;
					        	pets.add(new PET(olOut[0], olOut[1]));
					        	// break;
			        		}
			        	}
			        }
				}
			}
		} // --------------------------------------------------------- finding pets
    }
//*/
    public void readHeadId(int t, int fStart, int id) throws Exception {
    	int oCount = 0;
		System.err.println("readHeadId\t" + vName + "\t" + id);
    	if (read) {
    		if (t == OligoFile.F) 
    			openFileRead(vName + "/file" + fStart);
    		else 
    			openFileRead(vName + "/sort" + fStart);
    		dfn = fStart;
        	Oligo ol = new Oligo();
			idx = 0;
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				if (t == OligoFile.F) 
					readOligo(ol, "/file", Parameters.fSize);
				else
					readOligo(ol, "/sort", Parameters.fSortSize);
				if (ol.id1 == id)
					System.err.println(idx + "\t" + oCount++ + ":\t" + ol);
				idx++;
			}		
    	} else {
			System.err.println(" File open for write");
		}
		displayMeta();
   }
    public void ace() throws Exception {
		int fSize = 0;
		String line = "";
		System.err.println("vName:\t" + vName + "\trName:\t" + rName);
    	if (read) {
   			openFileRead(vName + "/file" + 1);
   	     	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/ace"));
			BufferedReader readerRef = new BufferedReader(new FileReader(rName));
			line = readerRef.readLine();
			StringTokenizer tokens = new StringTokenizer(line, ":");	
			if (tokens.countTokens() >= 6) {
				// tokens.nextToken();
				for(int i = 0; i < 5; i++)
					tokens.nextToken();
				fSize = new Integer(tokens.nextToken());
			}
			// System.err.println(line + "\t" + fSize);

   	     	writer.write("AS 1 " + numberOfOligos + "\n");
   	     	writer.write("CO " + rName + " " + fSize + " " + numberOfOligos + " U\n");
    		dfn = 1;
        	Oligo ol = new Oligo();
			idx = 0;
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				readOligo(ol, "/file", Parameters.fSize);
				writer.write(ol.start + "\t" + ol.toSeq() + "\n");
				idx++;
			}		
			readerRef.close();
			writer.close();
    	} else {
			System.err.println(vName + " File open for write");
		}
		// displayMeta();
   }

    /*
     	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/meta.txt"));
    	System.err.println(vName + "/meta.txt,\tminID\t" + (minID-1));
    	writer.write("Ref:\t" + rName); // numberOfIDs
    	writer.write("\nminID:\t"      + minID);
    	writer.write("\nnumberOfOligos:\t" + numberOfOligos);
    	writer.write("\nnumberOfUQ:\t" + numberOfUQ);
    	writer.write("\nnumberOfUD:\t" + numberOfUD);
    	writer.write("\nweightOfAT:\t" + weightOfAT);
    	writer.write("\nweightOfCG:\t" + weightOfCG);
    	writer.write("\ndfn:\t" + dfn);
    	writer.write("\nnumberOfBACs:\t" + numberOfBACs + "\n");
		for (int i = 0; i < numberOfBACs; i++) {
			writer.write(BACs[i] + "\n");
		}
    	writer.close();

     */
    //*/
    public void readHead(int last, int first) throws Exception {
    	if (read) {
   			openFileRead(vName + "/file" + 1);
    		dfn = 1;
        	Oligo ol = new Oligo();
			idx = 0;
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				readOligo(ol, "/file", Parameters.fSize);
				if ((idx >= first && idx < last) || last < 0)
					System.err.println(idx + ":\t" + ol);
				if (idx > last) 
					break;
				idx++;
			}		
    	} else {
			System.err.println(" File open for write");
		}
		// displayMeta();
   }
    
    public long available() {
    	return numberOfOligos - idx;
    }
    
    public boolean readOligo(Oligo ol) throws Exception {
    	boolean ret = true;
    	if (idx == 0) {
    		dfn = 1;
    		// System.err.println("Opening file:\t" + dfn);
			openFileRead(vName + "/file" + dfn);
    	}
    	if (idx < numberOfOligos) {
    		ret = readOligo(ol, "/file", Parameters.fSize);
    	} else {
    		ret = false;
    	}
    	idx++;
    	return ret;
    }
    public boolean readOligo(Oligo ol, String fil, int fsz) throws Exception {
    	boolean ret = true;
    	int fl = (int) (1 + oPointer / (long) fsz + (long) iDfn);
    	if (fl > dfn) {
  			closeFileRead();
			dfn++;
			try {
				// System.err.println("Opening file:\t" + dfn);
				openFileRead(vName + fil + dfn); // ----------------------------------------------------- "/file"
			} catch (Exception e) {
				System.err.println("Error opening file:\t" + dfn + "\t" + e);
			}
    	}
    	try {
        	ol.d[0]  = strIn.readLong();
        	ol.d[1]  = strIn.readLong();
        	ol.ch    = strIn.readByte();
        	ol.start = strIn.readInt();
        	ol.id1   = strIn.readInt();
        	ol.size  = strIn.readByte();
        	ol.flg   = strIn.readByte();
        	ol.u0    = strIn.readByte();
        	ol.weight = strIn.readByte();
    	} catch (Exception e) {
    		ret = false;
    		System.err.println("oPointer:\t" + oPointer + "\t" + e);
    	}
    	ol.weight = ol.weight / 100;
    	oPointer++;
    	return ret;
    }
    public void writeOligo(Oligo ol) throws Exception {
    	char bp0;
    	bp0 = ol.getBPat(0);
		if (ol.id1 < minID)
			minID = ol.id1;
    	if (bp0 == 'A' || bp0 == 'T') {
    		weightOfAT += ol.weight;
    	} else {
    		weightOfCG += ol.weight;
    	}
		if (ol.id1 < minID)
			minID = ol.id1;
		if (ol.isUQ())
			numberOfUQ += 1;
		else 
			numberOfUD += 1;
    	if (numberOfOligos % Parameters.fSize == 0) {
    		// if (numberOfOligos > 0)
    		if (dfn > iDfn) {
    			closeFileWrite();
    		}
			dfn++;
			openFileWrite(vName + "/file" + dfn);
    	}
    	strOut.writeLong(ol.d[0]);
    	strOut.writeLong(ol.d[1]);
    	strOut.writeByte(ol.ch); // int
    	strOut.writeInt(ol.start);
    	strOut.writeInt(ol.id1);
    	strOut.writeByte(ol.size); // int
    	strOut.writeByte(ol.flg); // int
   		strOut.writeByte(ol.u0); // int
    	strOut.writeByte((byte) ((100 * ol.weight) + 0.5)); // float
    	numberOfOligos++;
    }
    public void writeSOligo(Oligo ol) throws Exception {
    	if (numberOfOligos % Parameters.fSortSize == 0) {
    		if (dfns > iDfns) {
    			closeFileWrite();
    		}
			dfns++;
			openFileWrite(vName + "/sort" + dfns);
    	}
		if (ol.id1 < minID)
			minID = ol.id1;
    	strOut.writeLong(ol.d[0]);
    	strOut.writeLong(ol.d[1]);
    	strOut.writeByte(ol.ch); // int
    	strOut.writeInt(ol.start);
    	strOut.writeInt(ol.id1);
    	strOut.writeByte(ol.size); // int
    	strOut.writeByte(ol.flg); // int
   		strOut.writeByte(ol.u0); // int
    	strOut.writeByte((byte) ((100 * ol.weight) + 0.5)); // float
    	numberOfOligos++;
    }
    //*
    public void sort_WriteOligo(Oligo ol, int sortOn) throws Exception {
    	// System.err.print("writeRCOligo:\t");
    	char bp0;
    	bp0 = ol.getBPat(0);
		if (ol.id1 < minID)
			minID = ol.id1;
    	if (bp0 == 'A' || bp0 == 'T') {
    		weightOfAT += ol.weight;
    	} else {
    		weightOfCG += ol.weight;
    	}
		if (ol.isUQ())
			numberOfUQ += 1;
		else 
			numberOfUD += 1;
		bunk.add(ol);//(new Oligo(ol));
    	if (bunk.size() >= Parameters.fSortSize) {
    		saveSBunk(sortOn);
    	}
    }
    
    public void saveSBunk(int sIndex) throws Exception {
    	int bk = Oligo.sortIndex;
    	Oligo ol;
    	Oligo.sortIndex = sIndex;
		Collections.sort(bunk);
	    Iterator<Oligo> it = bunk.iterator();
        for (; it.hasNext(); ) {
        	ol = (Oligo) it.next ();
        	writeSOligo(ol);
        }
		bunk.clear();
		Oligo.sortIndex = bk;
    }
    //*/
    public void openFileWrite(String fl) throws Exception {
    	// System.err.println("openFileWrite" + fl);
    	ow = true;
		fw     = new FileOutputStream(fl);
	    bw     = new BufferedOutputStream(fw, buffSize);
	    strOut = new DataOutputStream(bw);	    		
    }
    private void closeFileWrite() throws Exception {
    	// System.err.print("x");
    	strOut.flush();
    	// System.err.print("=");
    	bw.flush();
    	// System.err.print("+");
    	fw.flush();
    	System.err.print(" *");
    	bw.flush();  	
    	strOut.flush();
    	
		strOut.close();    	
		bw.close();
		fw.close();
    }
    private void openFileRead(String fl) throws Exception {
		// System.err.println("\nSorting file " + dfn + "\nTiming:\tRead\tSort\tWrite\tFile");
    	if(numberOfOligos > 0) {
        	fi    = new FileInputStream(fl);
        	bi    = new BufferedInputStream(fi, buffSize);
    		strIn = new DataInputStream(bi);
    	}
    }
    private void closeFileRead() throws Exception {
		strIn.close();    	
		bi.close();
		fi.close();
    }

    public  void sortPET(String file) throws Exception {
    	String[] lst = {file + "0", file + "1"};
    	Oligo.sortIndex = Oligo.ID; // ??
    	join(lst, Oligo.ID);
    }
    public  void join(String[] lst, int sIndex) throws Exception {
    	boolean u0 = true;
    	int count = 0;
    	Oligo ol = new Oligo();
		bunk.clear();
    	for (int i = 0; i < lst.length; i++) {
    		OligoFile fl = new OligoFile(lst[i], "", OligoFile.R);
    		fl.dfn = iDfn + 1;
    		fl.openFileRead(fl.vName + "/file" + fl.dfn);
    		while(fl.strIn.available() > 0 || fl.dfn < numberOfFiles) {
        		fl.readOligo(ol, "/file", Parameters.fSize); // 332
        		if (i == 0)
        			u0 = true;
        		else
        			u0 = false;
        		ol.setU0(u0);
        		bunk.add(new Oligo(ol));
        		count++;
    	    	if (count >= Parameters.fSortSize) {
    	    		saveSBunk(sIndex);
    	    		count = 0;
    	    	}
    		}
    		fl.close();
    	}
		// Collections.sort(trg.bunk);
		saveSBunk(sIndex);
		numberOfFiles = dfn        = dfns;
    }
    
    public  void sortDB() throws Exception {
    	Oligo.sortIndex = Oligo.SEQ;
    	sort1_DB(vName);
    	trg  = new OligoFile(vName, "", OligoFile.R);
    	trg.joinSort();
    	trg.close();
    }
    public  void sort(String targetFile) throws Exception {
    	sort1(targetFile);
    	trg  = new OligoFile(targetFile, rName, OligoFile.R);
    	trg.joinSort();
    	trg.close();
    }
    private void saveBunk() throws Exception { // used by sort1
    	Oligo ol;
	    Iterator<Oligo> it = bunk.iterator();
        while (it.hasNext()) {
        	ol = (Oligo) it.next ();
        	trg.writeSOligo(ol);
        }
		bunk.clear();
		System.gc();
    }
//*
    private void sort1(String targetFile) throws Exception {
    	int count = 0;
    	Oligo ol = new Oligo();

    	trg = new OligoFile(targetFile, rName, OligoFile.W);
    	bunk.clear();
		dfn = iDfn + 1;
		openFileRead(vName + "/file" + dfn);
    	
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol, "/file", Parameters.fSize); // 332
	    	bunk.add(new Oligo(ol));
    		count++;
	    	if (count >= Parameters.fSortSize) {
	    		Collections.sort(bunk);
		    	saveBunk();
	    		count = 0;
	    	}
		}
		Collections.sort(bunk);
    	saveBunk();
    	trg.dfn        = trg.dfns;
    	trg.numberOfUD = numberOfUD;
    	trg.numberOfUQ = numberOfUQ;
    	trg.weightOfAT = weightOfAT;
    	trg.weightOfCG = weightOfCG;
    	trg.close();
    } //*/

    private void sort1_DB(String targetFile) throws Exception {
    	int count = 0;
    	Oligo ol = new Oligo();

    	trg = new OligoFile(targetFile, rName, OligoFile.W);
    	bunk.clear();
		dfn = iDfn + 1;
		openFileRead(vName + "/file" + dfn);
    	
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol, "/file", Parameters.fSize); // 332
    		if(ol.isRC())
    			ol.revers();
	    	bunk.add(new Oligo(ol));
    		count++;
	    	if (count >= Parameters.fSortSize) {
	    		Collections.sort(bunk);
		    	saveBunk();
	    		count = 0;
	    	}
		}
		if(count > 0) {
			Collections.sort(bunk);
			saveBunk();
		}
    	trg.dfn        = trg.dfns;
    	trg.numberOfUD = numberOfUD;
    	trg.numberOfUQ = numberOfUQ;
    	trg.weightOfAT = weightOfAT;
    	trg.weightOfCG = weightOfCG;
    	trg.close();
    } //*/
    private boolean getOligo(Oligo ol, int frm) throws Exception { // used by joinSort
    	boolean ret = true;
    	if (sstrIn[frm].available() > 0) {
        	ol.d[0]  = sstrIn[frm].readLong();
        	ol.d[1]  = sstrIn[frm].readLong();
        	ol.ch    = sstrIn[frm].readByte();
        	ol.start = sstrIn[frm].readInt();
        	ol.id1   = sstrIn[frm].readInt();
        	ol.size  = sstrIn[frm].readByte();
        	ol.flg   = sstrIn[frm].readByte();
       		ol.u0    = sstrIn[frm].readByte();
        	ol.weight = sstrIn[frm].readByte();
        	ol.weight = ol.weight / 100;
    	} else {
    		ret = false;
    	}
    	return ret;
    }
    public void joinFilterContaminationSort() throws Exception {
    	int noFiles;
    	int countNoise = 0;
    	boolean ntrue = true;
    	int cmp = 0;
    	int next = 0;
    	Oligo noise = new Oligo();
		String noiseDB = Parameters.pathData + "/contaminationDB";
		OligoFile ofNoise = new OligoFile(noiseDB, "", OligoFile.R);
		System.err.println("Contamination database size:\t" + ofNoise.numberOfOligos);
    	noFiles = numberOfFiles;
    	clearMeta();
    	System.err.println("noFiles:\t" + noFiles);
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft;
    	
    	sfi = new FileInputStream[noFiles]; 
    	sbi = new BufferedInputStream[noFiles];
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
    		oft       = new OligoFileTop();
    		sfi[i]    = new FileInputStream(vName + "/sort"+ (i+1));
    		sbi[i]    = new BufferedInputStream(sfi[i], 5000);
    		sstrIn[i] = new DataInputStream(sbi[i]);
    		getOligo(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		ntrue = ofNoise.readOligo(noise); // =================================================================
		// System.err.println(" ------------------- NOISE:\t" + noise);
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			if(ntrue) {
				cmp = noise.equal(oft.ol, oft.ol.size);
				if(cmp == 0) {
					oft.ol.setLQ(true);
					countNoise++;
					// System.err.println(" ------------------- NOISE:\t" + oft.ol);
				} else if(cmp > 0) {
					while (ntrue && cmp > 0) {
						ntrue = ofNoise.readOligo(noise);
						if (ntrue)
							cmp = noise.equal(oft.ol, oft.ol.size);
					}
				}
			}
			writeOligo(oft.ol);
			if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    		sbi[i].close();
    		sfi[i].close();
    	}
    	writeMeta();
    	ofNoise.close();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(vName + "/sort"+ (i+1))).delete();
    	}
    	System.err.println("countNoise:\t" + countNoise);
    }
    public void joinSort() throws Exception {
    	int noFiles;
    	int next = 0;
    	noFiles = numberOfFiles;
    	clearMeta();
    	System.err.println("noFiles:\t" + noFiles);
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft;
    	
    	sfi = new FileInputStream[noFiles]; 
    	sbi = new BufferedInputStream[noFiles];
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
    		oft       = new OligoFileTop();
    		sfi[i]    = new FileInputStream(vName + "/sort"+ (i+1));
    		sbi[i]    = new BufferedInputStream(sfi[i], 5000);
    		sstrIn[i] = new DataInputStream(sbi[i]);
    		getOligo(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			writeOligo(oft.ol);
			if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    		sbi[i].close();
    		sfi[i].close();
    	}
    	writeMeta();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(vName + "/sort"+ (i+1))).delete();
    	}
    }

    public void joinSortU2() throws Exception {
    	// int outCount = 0;
    	// Oligo last = new Oligo();
    	int u0;
    	Oligo ol1 = new Oligo(); 
    	l[1] = 0;
    	int noFiles;
    	int next = 0;
    	noFiles = numberOfFiles;

    	clearMeta();
    	// OligoFile out  = new OligoFile(vName + "/out", OligoFile.W);
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft;
    	
    	sfi = new FileInputStream[noFiles]; 
    	sbi = new BufferedInputStream[noFiles];
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
			System.err.println("sFile\t" + i);
    		oft       = new OligoFileTop();
    		sfi[i]    = new FileInputStream(vName + "/sort"+ (i+1));
    		sbi[i]    = new BufferedInputStream(sfi[i]);
    		sstrIn[i] = new DataInputStream(sbi[i]);
    		getOligo(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		if (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			ol1 = new Oligo(oft.ol);
			if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			l[0] = l[1];
			l[1] = oft.ol.matchR(ol1);
	        if (l[0] > l[1]) {
	        	u0 = l[0];
	        } else {
	        	u0 = l[1];
	        }
    		ol1.u0 = (byte) u0;
    		// ol1.u[1] = (byte) (u0 + 1);
    		
			writeOligo(ol1); 
			/*
			if (ol1.u0 > 100) {
				if (ol1.compareTo(last) != 0) {
					outCount++;
					out.writeOligo(ol1); 
					last = ol1;
				}
    		}
			//*/
	   		ol1 = new Oligo(oft.ol);
    		if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    		sbi[i].close();
    		sfi[i].close();
    	}
    	writeMeta();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(vName + "/sort"+ (i+1))).delete();
    	}
    	// out.close();
    }

    public  void fillU1(String targetFile) throws Exception {

     	Oligo ol0 = new Oligo();
    	Oligo ol1 = new Oligo();
    	trg = new OligoFile(targetFile, rName, OligoFile.W);
		dfn = iDfn + 1;
		// System.err.println("Opening file:\t" + dfn);
		openFileRead(vName + "/file" + dfn);
		readOligo(ol0); 
		ol0.u0 = 0;
		trg.writeOligo(ol0);
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol1); 
    		ol0.u0 = (byte) ol1.match(ol0);
    		trg.writeOligo(ol1);
    		ol0 = new Oligo(ol1);
		}
		trg.close();
    }

    public  void fillU2(String targetFile) throws Exception {

    	int u0 = 0;
    	Oligo ol0 = new Oligo();
    	Oligo ol1 = new Oligo();
    	l[1] = 0;
    	trg = new OligoFile(targetFile, rName, OligoFile.W);
		dfn = iDfn + 1;
		// System.err.println("Opening file:\t" + dfn);
		openFileRead(vName + "/file" + dfn);
		readOligo(ol0); // ------------------------------------------- , "/file", fSize);
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol1); // --------------------------------------- , "/file", fSize);
    		l[0] = l[1];
    		l[1] = ol1.match(ol0);
	        if (l[0] > l[1]) {
	        	u0 = l[0];
	        } else {
	        	u0 = l[1];
	        }
    		ol0.u0 = (byte) u0;
    		trg.writeOligo(ol0);
    		ol0 = new Oligo(ol1);
		}
		ol0.u0 = (byte) u0;
		trg.writeOligo(ol0);
		trg.close();
    }

    static public boolean delete(File path) {
        if(path.exists()) {
            File[] files = path.listFiles();
            for(int i=0; i<files.length; i++) {
               files[i].delete();
            }
        } else {
        	System.err.println("file not found:\t");
        }
        return(path.delete());
      }
    
}
