package JavaCode;

import java.util.ArrayList;

/*
SLIDER: Local Short Reads DNA Sequence Alignment
Beta Release; Ver 0.1, 22th Mar. 2008
By: Nawar Malhis
Canada's Michael Smith Genome Sciences Centre
*/

public class Seq {
	
    static final char[] nucl  = {'A', 'C', 'G', 'T'}; // used to convert int to char
	static final int       A = 0;          // 
	static final int       C = 1;          // 
	static final int       G = 2;          // 
	static final int       T = 3;          // 

	public int       size;
	// public int       nomm;
	public int        l;
	public float[][]  p;
	public byte[][][] q;
	public String     posibleSeq;
	public float      psp;
	public int[][]    alt;
	public int[]      used;  // the list of not good reads
	public byte[]     acc;
	public byte[]     count; // this value is the number of possible nucletids
	public int        totalUd;
	public float      sumw;
	public int        id;
	// public int       chr;
	// public int       side;
	
	private static Oligo oligo;

	public ArrayList<Oligo> bunk; 

	public Seq() {
		size  = Parameters.oMxSize;
		// nomm  = 0;
		p     = new float[size][4];
		q     = new byte [2][size][4];
		count = new byte [size];
		acc   = new byte [size];
		alt   = new int  [size][4];
		used  = new int  [size];
		
		bunk        = new ArrayList<Oligo>(); 
	}
	
	float getP_MPS(int[][] histo, float[][] histoW, float[][] cnt, int idxOne) { 
		int mx;
		double rt = 1;
		for(int i = 0; i < size; i++) {
			mx = 0;
			for (int j = 1; j < 4; j++) {
				if (p[i][mx] < p[i][j])
					mx = j;
			}
			rt = rt * p[i][mx];
			if (p[i][mx] < Parameters.seqSegNucP) {
				histoW[idxOne][i] += p[i][mx];
				histo[idxOne][i]++;
				cnt[idxOne][i] += count[i];
				// nomm++;
			}
		}
		psp = (float) rt;
		return (float) rt;
	}
	
	public void getMPS()  throws Exception {
    	char[] mps = posibleSeq.toCharArray();
		int mx;
		float rt = 1;
		bunk.clear();
		for(int i = 0; i < size; i++) {
			mx = 0;
			for (int j = 1; j < 4; j++) {
				if (p[i][mx] < p[i][j])
					mx = j;
			}
			mps[i] = nucl[mx];
			rt = rt * p[i][mx];
			/*
			if (p[i][mx] < Parameters.seqSegNucP) {
				nomm++;
			}
			//*/
		}
		posibleSeq = new String(mps);
		oligo = new Oligo(posibleSeq);
		oligo.weight = rt;
		bunk.add(new Oligo(oligo));
	}
	
	public int solvePrbLine(int ll) {
		char[] ps = new char[size];
		int x;
		l = ll;
		totalUd = 0;
		for (int i = 0; i < size; i++) {
			ps[i] = 0;
			count[i] = 0;
			used[i]  = -1;
			for (int j = 0; j < 4; j++) {
				alt[i][j] = -1;
				x = q[l][i][j]; //
				p[i][j] = (float) (1/(1+Math.pow(10,(((float)-x)/10))));
				if (p[i][j] > (1-Parameters.seqSegNucP)) {
					if(ps[i] == 0) {
						ps[i] = nucl[j];
					}
					if (p[i][j] < Parameters.seqSegNucP) {
						used[totalUd]  = i;
					}
					alt[i][count[i]] = j;
					count[i]++;
				}
			}
			if (count[i] > 0)
				ps[i] = nucl[alt[i][count[i]-1]];
			if (used[totalUd] > -1) {
				totalUd++;
			}
		}
		posibleSeq = new String(ps).substring(0, size); 
		int ret = 4;
		if (totalUd == 0) {
			ret = 0;
		} else if (totalUd <= Parameters.maxUDN) {
			ret = 1;
		} else {
			// must set size & totalUd
			size = used[Parameters.maxUDN];
			if (size >= Parameters.oMxSize) {
				ret = 1;
				totalUd = Parameters.maxUDN;
			} else {
				size = Parameters.oMxSize;
				ret = 4;
			}
		}
		// System.err.println("totalUd:\t" + totalUd + ",\t" + size);
		// System.err.println("ps:\t" + ps + ",\t" + size);
		posibleSeq = new String(ps).substring(0, size); 
		return ret;
	}

	public int getPS() throws Exception {
		int t1;
		int xx = 1;
		sumw = 0;
    	char[] olc = posibleSeq.toCharArray();
    	int[] olcAlt = new int[posibleSeq.length()];
   		for (int i = 0; i < size; i++) {
    		olcAlt[i] = count[i] - 1;
    	}
   		bunk.clear();
    	while(xx == 1) {
    		xx = 0;
    		oligo = new Oligo(posibleSeq);
    		psp = 1;
    		for (int i = 0; i < size; i++) {
    			psp = psp * p[i][alt[i][olcAlt[i]]];
    		}
    		oligo.weight = psp;
    		sumw += psp;
    		if (psp > Parameters.minWeight) {
        		bunk.add(new Oligo(oligo));
    		}
    		for (int i = totalUd - 1; i >= 0; i--) {
    			if (olcAlt[used[i]] == 0) {
    				olcAlt[used[i]] = count[used[i]] - 1;
    				t1 = olcAlt[used[i]];
    				olc[used[i]] = nucl[alt[used[i]][t1]];
    			} else {
    				olcAlt[used[i]]--;
    				t1 = olcAlt[used[i]];
    				olc[used[i]] = nucl[alt[used[i]][t1]];
    				posibleSeq = new String(olc);
    				xx = 1;
    				break;
    			}
    		}
    
    	}
    	return -1;
	}

	public boolean isCrisp(int i) {
		for (int j = 0; j < 4; j++) {
			if (p[i][j] > Parameters.seqSegNucP)
				return true;
		}
		return false;
	}
	public void display() throws Exception {
		System.out.println(posibleSeq + "\t" + trim(psp));
		for (int j = 0; j < 4; j++) {
			for (int i = 0; i < size; i++) {
				System.out.print(trim(p[i][j]) + "\t");
			}
			System.out.println();
		}
		System.out.println("--------------------------");
		for (int i = 0; i < size; i++) {
			System.out.print(i + "\t");
		}
		System.out.println();
		System.out.println("Used:\t" + totalUd);
		for (int i = 0; i < size; i++) {
			System.out.print(used[i] + "\t");
		}
		System.out.println();
		System.out.println("Count:");
		for (int i = 0; i < size; i++) {
			System.out.print(count[i] + "\t");
		}
		System.out.println();
		System.out.println("Alt:");
		for (int j = 0; j < 4; j++) {
			for (int i = 0; i < size; i++) {
				System.out.print(alt[i][j] + "\t");
			}
			System.out.println();
		}
		System.err.println((new Oligo(posibleSeq))); //.display();
	}
	
	private float trim(float f) {
		return ((float) ((int) (f * 1000))) / (float) 1000.0; 
	}

	public int  getBpInt(char b) {
    	if (b == 'A')
    	    return A;
    	if (b == 'C')
    	    return C;
    	if (b == 'G')
    	    return G;
    	if (b == 'T')
    	    return T;
    	return -1;
        }
}
